"""Helper functions for Item Collection"""
import json
from flask import Response, jsonify, request, abort
from hydra_python_core.doc_writer import HydraStatus, HydraError
from hydrus.data import crud
from hydrus.data.exceptions import (
    ClassNotFound,
    InstanceExists,
    PropertyNotFound,
    InstanceNotFound
)
from hydrus.helpers import (
    set_response_headers,
    checkClassOp,
    validObjectList,
    type_match,
    check_required_props,
    send_sync_update,
    get_link_props_for_multiple_objects,
    get_iri_from_int_list,
    error_response,
    getType,
)
from hydrus.utils import (
    get_session,
    get_api_name,
    get_hydrus_server_url,
    get_collections_and_parsed_classes
)
from hydrus.socketio_factory import socketio


def items_put_response(path: str, int_list="") -> Response:
    """
    Handles PUT operation to insert multiple items.

    :param path: Path for Item Collection
    :type path: str
    :param int_list: Optional String containing ',' separated ID's
    :type int_list: List
    :return: Appropriate response for the PUT operation on multiple items.
    :rtype: Response
    """
    object_ = json.loads(request.data.decode('utf-8'))
    object_ = object_["data"]
    _, parsed_classes = get_collections_and_parsed_classes()
    if path in parsed_classes:
        class_path = path
        obj_type = getType(path, "PUT")
        incomplete_objects = []
        for obj in object_:
            if not check_required_props(class_path, obj):
                incomplete_objects.append(obj)
                object_.remove(obj)
        link_props_list, link_type_check = get_link_props_for_multiple_objects(class_path,
                                                                               object_)
        if validObjectList(object_) and link_type_check:
            type_result = type_match(object_, obj_type)
            # If Item in request's JSON is a valid object
            # ie. @type is one of the keys in object_
            if type_result:
                # If the right Item type is being added to the
                # collection
                try:
                    # Insert object and return location in Header
                    object_id = crud.insert_multiple(
                        objects_=object_, session=get_session(), id_=int_list)
                    headers_ = [{"Location": f"{get_hydrus_server_url()}"
                                             f"{get_api_name()}/{path}/{object_id}"}]
                    if len(incomplete_objects) > 0:
                        status = HydraStatus(code=202,
                                             title="Object(s) missing required property")
                        response = status.generate()
                        response["objects"] = incomplete_objects
                        return set_response_headers(
                            jsonify(response), headers=headers_,
                            status_code=status.code)
                    else:
                        path_url = f"{get_hydrus_server_url()}{get_api_name()}/{path}"
                        response_url_list = get_iri_from_int_list(path_url, object_id)
                        status_description = f"Objects with ID {object_id} successfully added"
                        status = HydraStatus(code=201, title="Objects successfully added",
                                             desc=status_description)
                        status_response = status.generate()
                        status_response["iri"] = response_url_list
                        return set_response_headers(
                            jsonify(status_response), headers=headers_,
                            status_code=status.code)
                except (ClassNotFound, InstanceExists, PropertyNotFound) as e:
                    error = e.get_HTTP()
                    return error_response(error)

        error = HydraError(code=400, title="Data is not valid")
        return error_response(error)


def items_delete_response(path: str, int_list="") -> Response:
    """
    Handles DELETE operation to insert multiple items.

    :param path: Path for Item Collection
    :type path: str
    :param int_list: Optional String containing ',' separated ID's
    :type int_list: List
    :return: Appropriate response for the DELETE operation on multiple items.
    :rtype: Response
    """
    _, parsed_classes = get_collections_and_parsed_classes()
    if path in parsed_classes:
        class_type = getType(path, "DELETE")

    if checkClassOp(class_type, "DELETE"):
        # Check if class_type supports PUT operation
        try:
            # Delete the Item with ID == id_
            crud.delete_multiple(int_list, class_type, session=get_session())
            method = "DELETE"
            path_url = f"{get_hydrus_server_url()}{get_api_name()}/{path}"
            last_job_id = crud.get_last_modification_job_id(session=get_session())
            id_list = int_list.split(',')
            for item in id_list:
                resource_url = path_url + item
                new_job_id = crud.insert_modification_record(method,
                                                             resource_url,
                                                             session=get_session())
                send_sync_update(socketio=socketio, new_job_id=new_job_id,
                                 last_job_id=last_job_id, method=method,
                                 resource_url=resource_url)
                last_job_id = new_job_id
            status_description = f"Objects with ID {id_list} successfully deleted"
            status = HydraStatus(code=200,
                                 title="Objects successfully deleted",
                                 desc=status_description)
            return set_response_headers(jsonify(status.generate()))

        except (ClassNotFound, InstanceNotFound) as e:
            error = e.get_HTTP()
            return error_response(error)

    abort(405)



"""
This file (test_crud.py) contains the unit tests for
CRUD operations in hydrus.data.crud.
"""
import random
import uuid

from hydra_python_core.doc_writer import HydraLink, DocUrl

import hydrus.data.crud as crud
from hydrus.data.exceptions import PropertyNotGiven
from hydrus.tests.conftest import gen_dummy_object


def test_crud_insert_response_is_str(drone_doc_parsed_classes, drone_doc, session,
                                     init_db_for_crud_tests):
    """Test CRUD insert response is string"""
    object_ = gen_dummy_object(random.choice(drone_doc_parsed_classes), drone_doc)
    id_ = str(uuid.uuid4())
    response = crud.insert(object_=object_, id_=id_, session=session)

    assert isinstance(response, str)


def test_crud_get_returns_correct_object(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD get returns correct object"""
    object_ = gen_dummy_object(random.choice(drone_doc_parsed_classes), drone_doc)
    id_ = str(uuid.uuid4())
    response = crud.insert(object_=object_, id_=id_, session=session)
    object_ = crud.get(id_=id_, type_=object_['@type'], session=session, api_name='api')
    assert isinstance(response, str)
    assert object_['@id'].split('/')[-1] == id_


def test_get_for_nested_obj(drone_doc_parsed_classes, drone_doc, session, constants):
    """Test CRUD get operation for object that can contain other objects."""
    expanded_base_url = DocUrl.doc_url
    for class_ in drone_doc_parsed_classes:
        for prop in drone_doc.parsed_classes[class_]['class'].supportedProperty:
            if not isinstance(prop.prop, HydraLink):
                if expanded_base_url in prop.prop:
                    dummy_obj = gen_dummy_object(class_, drone_doc)
                    nested_class = prop.prop.split(expanded_base_url)[1]
                    obj_id = str(uuid.uuid4())
                    response = crud.insert(object_=dummy_obj, id_=obj_id, session=session)
                    object_ = crud.get(id_=obj_id, type_=class_, session=session,
                                       api_name='api')
                    assert prop.title in object_
                    nested_obj_id = object_[prop.title]
                    nested_obj = crud.get(id_=nested_obj_id, type_=nested_class,
                                          session=session, api_name='api')
                    assert nested_obj['@id'].split('/')[-1] == nested_obj_id
                    break


def test_searching_over_collection_elements(drone_doc_parsed_classes, drone_doc, session):
    """Test searching over collection elements."""
    expanded_base_url = DocUrl.doc_url
    for class_ in drone_doc_parsed_classes:
        target_property_1 = ''
        target_property_2 = ''
        for prop in drone_doc.parsed_classes[class_]['class'].supportedProperty:
            if isinstance(prop.prop, HydraLink):
                continue
            # Find nested object so we can test searching of elements by
            # properties of nested objects.
            if expanded_base_url in prop.prop:
                object_ = gen_dummy_object(class_, drone_doc)
                # Setting property of a nested object as target
                for property_ in object_[prop.title]:
                    if property_ != '@type':
                        object_[prop.title][property_] = 'target_1'
                        target_property_1 = '{}[{}]'.format(
                            prop.title, property_)
                        break
                break
            elif target_property_1 is not '':
                for property_ in object_:
                    if property_ != '@type':
                        object_[property_] = 'target_2'
                        target_property_2 = property_
                        break
                break

            if target_property_1 is not '' and target_property_2 is not '':
                # Set search parameters
                search_params = {
                    target_property_1: 'target_1',
                    target_property_2: 'target_2'
                }

                obj_id = str(uuid.uuid4())
                response = crud.insert(
                    object_=object_, id_=obj_id, session=session)
                search_result = crud.get_collection(API_NAME='api', type_=class_,
                                                    session=session, paginate=True,
                                                    page_size=5, search_params=search_params)
                assert len(search_result['members']) > 0
                search_item_id = search_result['members'][0]['@id'].split(
                    '/')[-1]
                assert search_item_id == obj_id
                break


def test_update_on_object(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD update on object"""
    random_class = random.choice(drone_doc_parsed_classes)
    object_ = gen_dummy_object(random_class, drone_doc)
    new_object = gen_dummy_object(random_class, drone_doc)
    id_ = str(uuid.uuid4())
    insert_response = crud.insert(object_=object_, id_=id_, session=session)
    update_response = crud.update(
        id_=id_,
        type_=object_['@type'],
        object_=new_object,
        session=session,
        api_name='api')
    test_object = crud.get(id_=id_, type_=object_['@type'], session=session, api_name='api')
    assert isinstance(insert_response, str)
    assert isinstance(update_response, str)
    assert insert_response == update_response
    assert test_object['@id'].split('/')[-1] == id_


def test_delete_on_object(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD delete on object"""
    object_ = gen_dummy_object(random.choice(drone_doc_parsed_classes), drone_doc)
    id_ = str(uuid.uuid4())
    insert_response = crud.insert(object_=object_, id_=id_, session=session)
    assert isinstance(insert_response, str)
    delete_response = crud.delete(id_=id_, type_=object_['@type'], session=session)
    response_code = None
    try:
        get_response = crud.get(
            id_=id_,
            type_=object_['@type'],
            session=session,
            api_name='api')
    except Exception as e:
        error = e.get_HTTP()
        response_code = error.code
    assert 404 == response_code


def test_get_on_wrong_id(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD get when wrong/undefined ID is given."""
    id_ = str(uuid.uuid4())
    type_ = random.choice(drone_doc_parsed_classes)
    response_code = None
    try:
        get_response = crud.get(id_=id_, type_=type_, session=session, api_name='api')
    except Exception as e:
        error = e.get_HTTP()
        response_code = error.code
    assert 404 == response_code


def test_delete_on_wrong_id(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD delete when wrong/undefined ID is given."""
    object_ = gen_dummy_object(random.choice(
        drone_doc_parsed_classes), drone_doc)
    id_ = str(uuid.uuid4())
    insert_response = crud.insert(object_=object_, id_=id_, session=session)
    response_code = None
    try:
        delete_response = crud.delete(id_=999, type_=object_['@type'], session=session)
    except Exception as e:
        error = e.get_HTTP()
        response_code = error.code
    assert 404 == response_code
    assert isinstance(insert_response, str)
    assert insert_response == id_


def test_insert_used_id(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD insert when used ID is given."""
    object_ = gen_dummy_object(random.choice(drone_doc_parsed_classes), drone_doc)
    id_ = str(uuid.uuid4())
    insert_response = crud.insert(object_=object_, id_=id_, session=session)
    response_code = None
    try:
        insert_response = crud.insert(
            object_=object_, id_=id_, session=session)
    except Exception as e:
        error = e.get_HTTP()
        response_code = error.code
    assert 400 == response_code


def test_get_on_wrong_type(session):
    """Test CRUD get when wrong/undefined class is given."""
    id_ = str(uuid.uuid4())
    type_ = 'otherClass'
    response_code = None
    try:
        get_response = crud.get(id_=id_, type_=type_, session=session, api_name='api')
    except Exception as e:
        error = e.get_HTTP()
    assert 400 == error.code


def test_delete_on_wrong_type(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD delete when wrong/undefined class is given."""
    object_ = gen_dummy_object(random.choice(drone_doc_parsed_classes), drone_doc)
    id_ = str(uuid.uuid4())
    insert_response = crud.insert(object_=object_, id_=id_, session=session)
    assert isinstance(insert_response, str)
    assert insert_response == id_
    response_code = None
    try:
        delete_response = crud.delete(id_=id_, type_='otherClass', session=session)
    except Exception as e:
        error = e.get_HTTP()
        response_code = error.code
    assert 400 == response_code


def test_insert_on_wrong_type(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD insert when wrong/undefined class is given."""
    object_ = gen_dummy_object(random.choice(drone_doc_parsed_classes), drone_doc)
    id_ = str(uuid.uuid4())
    object_['@type'] = 'otherClass'
    response_code = None
    try:
        insert_response = crud.insert(object_=object_, id_=id_, session=session)
    except Exception as e:
        error = e.get_HTTP()
        response_code = error.code
    assert 400 == response_code


def test_insert_multiple_id(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD insert when multiple ID's are given """
    objects = list()
    ids = '{},{}'.format(str(uuid.uuid4()), str(uuid.uuid4()))
    ids_list = ids.split(',')
    for index in range(len(ids_list)):
        object = gen_dummy_object(random.choice(drone_doc_parsed_classes), drone_doc)
        objects.append(object)
    insert_response = crud.insert_multiple(objects_=objects, session=session, id_=ids)
    for id_ in ids_list:
        assert id_ in insert_response


def test_delete_multiple_id(drone_doc_parsed_classes, drone_doc, session):
    """Test CRUD insert when multiple ID's are given """
    objects = list()
    ids = '{},{}'.format(str(uuid.uuid4()), str(uuid.uuid4()))
    random_class = random.choice(drone_doc_parsed_classes)
    for index in range(len(ids.split(','))):
        object = gen_dummy_object(random_class, drone_doc)
        objects.append(object)
    insert_response = crud.insert_multiple(objects_=objects, session=session, id_=ids)
    delete_response = crud.delete_multiple(id_=ids, type_=random_class, session=session)

    response_code = None
    id_list = ids.split(',')
    try:
        for index in range(len(id_list)):
            get_response = crud.get(
                id_=id_list[index],
                type_=objects[index]['@type'],
                session=session,
                api_name='api')
    except Exception as e:
        error = e.get_HTTP()
        response_code = error.code
    assert 404 == response_code


def test_insert_when_property_not_given(drone_doc_parsed_classes, drone_doc,
                                        session, constants):
    """Test CRUD insert operation when a required foreign key
    property of that resource(column in the table) not given"""
    expanded_base_url = DocUrl.doc_url
    for class_ in drone_doc_parsed_classes:
        for prop in drone_doc.parsed_classes[class_]['class'].supportedProperty:
            if not isinstance(prop.prop, HydraLink) and expanded_base_url in prop.prop:
                dummy_obj = gen_dummy_object(class_, drone_doc)
                nested_prop_title = prop.title
                continue
    # remove the foreign key resource on purpose for testing
    dummy_obj.pop(nested_prop_title)
    id_ = str(uuid.uuid4())
    try:
        insert_response = crud.insert(object_=dummy_obj, id_=id_, session=session)
    except PropertyNotGiven as e:
        error = e.get_HTTP()
        response_code = error.code
        assert 400 == response_code
